@@ -5,6 +5,11 @@ from uuid import uuid4
 from pathlib import Path
 from itertools import zip_longest
 
+try:
+    from contextlib import chdir
+except ImportError:
+    from contextlib_chdir import chdir
+
 from programlib import Agent, language_, Terminal
 
 def correctness(expected_outputs, outputs):
@@ -51,12 +56,13 @@ class Program():
             self.workdir = Path(__file__).parent / 'programs'
             self.ephemeral = True
 
-        if source:
-            self.language.write_source(self.workdir, self.name, source)
-        else:
-            source = self.language.read_source(self.workdir, self.name)
+        with chdir(self.workdir):
+            if source:
+                self.language.write_source(self.name, source)
+            else:
+                source = self.language.read_source(self.name)
 
-        self.stdout, self.exitstatus = self.language.build(self.workdir, self.name)
+            self.stdout, self.exitstatus = self.language.build(self.name)
 
         # Please send your opinions on 'not not' vs 'bool()' to dont@vadim.me
         self.compile_error = not not self.exitstatus
@@ -66,8 +72,9 @@ class Program():
 
     def read(self):
         """Obtain the source code of the program"""
-
-        return self.language.read_source(self.workdir, self.name)
+        
+        with chdir(self.workdir):
+            return self.language.read_source(self.name)
 
     def run(self, input_lines=[], force=True):
         """
@@ -81,7 +88,13 @@ class Program():
         Raw results will always be stored in `program.stdout` and `program.exitstatus` attributes
         """
 
-        self.stdout, self.exitstatus = self.language.run(self.workdir, self.name, input_lines)
+        try:
+            with chdir(self.workdir):
+                self.stdout, self.exitstatus = self.language.run(self.name, input_lines)
+        except RuntimeError as e:
+            self.exitstatus = 1
+            self.stdout += str(e)
+
         assert force or not self.exitstatus, f'Exit status {self.exitstatus}'
         return self.term.emulate(self.stdout)
     
@@ -90,34 +103,35 @@ class Program():
         Launch the program and get a process object to communicate with it interactively
         """
 
-        process = self.language.spawn(self.workdir, self.name)
+        with chdir(self.workdir):
+            process = self.language.spawn(self.name)
         return Agent(self, process, delimiter=delimiter)
 
-    def test(self, test_cases, force=True):
+    def test(self, test_cases):
         """
         Test the program against a list of input output pairs
 
-        If 'force=True', program outputs will be checked even if the exit status is non-zero
         If 'cache=True', all the test logs will be stored in `program.test_runs` attribute
         """
 
         test_runs = []
 
         for input_lines, expected_output_lines in test_cases:
-            test_run = None
-
-            if not self.compile_error:
-                try:
-                    output_lines = self.run(input_lines, force=force)
-                    test_run = TestRun(input_lines, expected_output_lines, 
-                                       output_lines, self.exitstatus,
-                                       correctness(expected_output_lines, output_lines))          
-                except AssertionError:
-                    pass
-
-            if not test_run:
+            if self.compile_error:
+                # The program is not runnable
                 test_run = TestRun(input_lines, expected_output_lines, 
-                                   [], self.exitstatus, 0)
+                                   self.term.emulate(self.stdout) if self.stdout else [], 
+                                   self.exitstatus, 0)
+            else:
+                output_lines = self.run(input_lines, force=True)
+                if self.exitstatus:
+                    # The program crashed
+                    corr = 0
+                else:
+                    # The program ran successfully
+                    corr = correctness(expected_output_lines, output_lines)
+                test_run = TestRun(input_lines, expected_output_lines, 
+                                   output_lines, self.exitstatus, corr) 
 
             test_runs.append(test_run)
 
@@ -127,8 +141,12 @@ class Program():
         return test_runs
 
     def save(self, path):
-        self.language.copy_source(self.workdir, self.name, path)
+        path = Path(path).absolute()
+
+        with chdir(self.workdir):
+            self.language.copy_source(self.name, path)
 
     def __del__(self):
         if self.ephemeral:
-            self.language.cleanup(self.workdir, self.name)
\ No newline at end of file
+            with chdir(self.workdir):
+                self.language.cleanup(self.name)
\ No newline at end of file
